/** The Wobzilla Kernel.

<p>
When wobzilla starts, an instance of WobzillaEngine is created.  In
each wobzilla application there is only one instance of WobzillaEngine. The
WobzillaEngine-object manages the main-document-window, the debugging
window and the menu-functions for saving and loading documents.  
</p>

<p>Upon document-loading the configuration of the editor is loaded
and WobzillaEngine instanciates the root-view that is specified in the
configuration and delegates processing to the root-view. It passes the
root-node of the data-document as model-node and the main-document as
output-node to the root-view (See <a
href="../wobzilla.html" target="_top">Wobzilla-Views</a> for more info on
views)</p>

<p>WobzillaEngine provides a lot of services to the views. Here is a list of all
services:
</p>

@provides wobzilla_transaction
@provides wobzilla_componentFactory
@provides wobzilla_dialogFactory
@provides wobzilla_popupFactory
@provides wobzilla_selectionContext
@provides wobzilla_xmlloader
@provides wobzilla_parser
@provides wobzilla_serializer

  @constructor

*/
function WobzillaEngine() {
    function dummy(){};
}

/** Transaction Handling in wobzilla. This service manages all changes to
    the data-model and can be used
   <ol>
     <li>to encapsulate different elementary operations in one atomar transaction. 
     Typically, this are operations that occur in the same user-interaction. A move, for
     example, consist of a remove and a insert operation.
     </li>
     <li>
     to undo of one or more transaction. Redo of the undone transactions is also possible.
     </li>
     <li>
     to notify other views, after a successful transaction, or after an undo transaction.
     This gives the views the oppertunity to refresh their state.
     </li>
   </ol>
   
    <p>Assigned to every transaction is a <a
    href="../wobzilla-framework.html#changeLog">wobzilla.changeLog</a>-object, that
    keeps the elementary node-operations in the order of their
    execution. This changelog is passed to the requery methods of the views, after 
    a successful transaction.
    </p>
    <p>The inverse transaction is determined with the reversion of the order 
    of the elementary operations and the inversion of each elementary operation.
    The inverses of the operations are calculated with the following table:</p>
    
     <table>
      <thead><th>operation</th><th>inverse operation</th><th>Attribute-transformations</th></thead>
      <tr><td><strong>DOMNodeInserted</strong></td><td><strong>DOMNodeRemoved</strong></td><td>inverse.removeBefore=insertBefore</td></tr>
      <tr><td><strong>DOMNodeRemoved</strong></td><td><strong>DOMNodeInserted</strong></td><td>inverse.insertBefore=removeBefore</td></tr>
      <tr><td><strong>DOMAttrChanged</strong></td><td><strong>DOMAttrChanged</strong></td><td>inverse.newValue=oldValue; inverse.oldValue=newValue</td></tr>
    </table>

  @constructor
 */
function wobzilla_transaction(INTERFACE_ONLY_NO_CONSTRUCTOR) {
    /** Starts a new transaction and clears the <em>redo</em>-stack.*/
    this.start = function();
    /** Ends a started transaction and pushes the transaction to the <em>undo</em>-stack.
        Calls the <code>requery</code>-method on the root-view and passes the <code>changeLog</code> of the transaction as a parameter.*/
    this.commit = function();
    /** removes a transaction from the <em>undo</em>-stack and inserts it into the <em>redo</em>-stack. Executes the <strong>inverse</strong> transaction and notifies the root-view.*/
    this.undo = function();
    /** removes a transaction from the <em>redo</em>-stack and inserts it into the <em>undo</em>-stack. Executes the transaction and notifies the root-view.*/
    this.redo = function();
}


/** <p>Given a configuration and a context and a set of services, a <code>ComponentFactory</code> finds an applicable componenent and instanciates it.
</p>
  <p>The components class can be specified in the <code>class</code>-attribute of the 
  root-configuration element. If no such attribute exists, the component-factory can
  try to decide the components class by evaluting the configuration or the context.
  </p>

  <p><code>WobzillaEngine</code>, for instance, can be provided with a
  shorcut-list configured in <code>wobzillaConf.xml</code>. Under
  <code>shortcuts</code> you'll find a mapping between
  configuration-names and class-names. This information will be used
  by the <code>wobzilla_componentFactory</code> provided by <code>WobzillaEngine</code>.
  </p>

  @constructor
 */
function wobzilla_componentFactory(INTERFACE_ONLY_NO_CONSTRUCTOR) {
    /** Finds a component for the given configuration and context,
        creates an uninitialized instance of that component 
        and calls <code>setup</code>.
        @param componentConfig the configuration node for the new component instance.
        @param logger the logger that should be used by the component instance.
        @param context the context that is available for the component instance.
        @param serviceManager the serviceManager, that provides a list of all services that are exposed to the component instance.
        @throws ComponentNotFound if the component class could not be found. The cause can be a missing <code>class</code>-attribue or shortcut, or that the component-code can't be loaded.
        @throws ConfigurationException if the component could not be configured due to an 
        invalid configuration. The component decides wether the configuration is valid or not.
        @throws ServiceException if the component requests a service that is not provided by the passed serviceManager
        @throws ContextException if the component requests a context-entry that is not provided by the passed context.
        
        @return an initialized component instance.
     */
    this.createComponent = function(componentConfig,logger,context,serviceManager);
    /** calls all the lifecycle methods: logEnable, contextualize, configure and service 
        ,in the stated oder, on an uninitialized component instance.
        @return an initialized component instance.
     */
    this.setup = function(componentInstance, config, logger,context,serviceManager);
}

/**
  @constructor
 */
function wobzilla_xmlloader(INTERFACE_ONLY_NO_CONSTRUCTOR) {
    /** Loads a xml-file and creates a document-object from it.
        The File must end with .xml to trigger correct xml-parsing 
        (XMLHttpRequest can be used for arbitrary files)

        TODO: currently a random number is passed as an additional request parameter, to avid some caching problems. Must find a better way todo this.
        @return a reference to the document-object.
    */
    this.load = function(filename);
}

/** 
  The selectionContext allows selection-handling accross
  component-borders.  Use <code>wobzilla_selectionContext</code> to
  communicate the selection events of on component to all registered
  components.  Example: If you have two list-components and want only
  one item selected in both lists, you could check for the <code>target</code>
  parameter to ensure that an selected entry in a list should be unselected if 
  the selection-event is targeted to the other list.
  @constructor
 */
function wobzilla_selectionContext(INTERFACE_ONLY_NO_CONSTRUCTOR) {
    /** Listeners must provide a selectionChanged method. The <code>target</code>
        and the <code>args</code> parameter of the <code>select</code> method are
        passed as parameters to the selectionChangedMethod.
        @see #select
    */
    this.addSelectionListener = function( selectionChangedMethod );
    /** Iterates through all registered listeners 
        and calls the <code>selectionChangedMethod</code> with the 
        <code>target</code> and <code>args</code> parameter.
    */
    this.select = function(target,args);
}


/**
   Use the dialogFactory to load dialog-files, create the dialog from
   the file-discription.
   @constructor
*/
function wobzilla_dialogFactory() {
    /** creates a new dialog.
     * @param dialogFile the path of the dialog file
     * @param titleText 
     * @param openMethod this method is called when the dialog is visible
     * @param closeMethod this method is called when the dialog is closed
     * @param left
     * @param top
     * @param width
     * @param height
     * @return the new dialog object
     */
    this.createDialog = function (dialogFile,titleText,openMethod,closeMethod,left,top,width,height);
    /** closes and destroys the dialog object*/
    this.closeDialog = function(dialog);
}

/**
   Parses a string and returns a xml node.
   @constructor
*/
function wobzilla_parser() {
    this.parseString = function( string );
}

/**
   Serializes a DOM-Tree fragment into a string.
   @constructor
*/
function wobzilla_serializer() {
    this.serializeToString = function( content);
}
