/* 
   Provides static util functions. Most of the functions are
   dealing with DOM-Nodes, but there is also a list-diff and access
   to a static logger-object.
*/
ELEMENT_NODE = 1;
ATTRIBUTE_NODE = 2;
TEXT_NODE = 3;
DOCUMENT_NODE = 9;

var WBZ_Serializer = new XMLSerializer();
var WBZ_DOMParser = new DOMParser();

/** serializes a document node to a text string and handles the proper encodings.
  @param node the node to be serialized
  @return the serialized string
 */
function WBZ_serializeXML(node)
{
    if (node == null) {
        return "";
    }
    return WBZ_Serializer.serializeToString(node);
}

/** parses an xml-fragment from a text.
    @param text the text that is parsed
    @param document a document object
    @return the node that is parsed
*/
function WBZ_parseXML(text,document) {
    var content = "<?xml version='1.0'?><wrapper>" + text +"</wrapper>";
    content = WBZ_DOMParser.parseFromString(content,"text/xml").firstChild;
    if (content.nodeName == "parsererror") 
        throw new Error(content.firstChild.nodeValue);

    return content.firstChild.cloneNode(true);
}

/** returns the first following sibling that is an element-node or null if no
 * such sibling exists. See DOM Specification for Definition of
 * nextSibling.*/
function WBZ_nextElementSibling(element) {
    while (element.nextSibling != null ) {
        element = element.nextSibling;
        if (element.nodeType == ELEMENT_NODE)
            return element;
    }
    return null;
}

/** returns the first previous sibling that is an element-node or null if no
 * such sibling exists. See DOM Specification for Definition of
 * previousSibling.*/
function WBZ_previousElementSibling(element) {
    while (element.previousSibling != null ) {
        element = element.previousSibling;
        if (element.nodeType == ELEMENT_NODE)
            return element;
    }
    return null;
}

/** returns the attribute value of an element. If the attribute is not set 
    the default value will be used.
 */
function WBZ_attribute(element,attributeName,defaultValue) {
    var value = element.getAttribute(attributeName);
    if (value == null)
        return defaultValue;
    else 
        return value;
}
/** Wraps multiple view-configurations and/or elements without a without a
    wobzilla-namespace in a WbzContainer config. See WbzContainer documentation 
    for more information.
 @see WbzContainer
 @return the configuration of a wrapper object
*/
function WBZ_wrapChildConfigs(childs) {
    var countViews = 0;
    var childView = null;
    for (var i = 0;i<childs.length; i++) {
        if (childs[i].nodeType == ELEMENT_NODE) {
            if (childs[i].namespaceURI == WOBZILLA_NS) {
                childView = childs[i];
                countViews ++;
            } else {
                countViews = 100; 
                // This is not a view element so we need to build a wrapper
            }
            // check if there is a non whitespace text-Node
        } else if (childs[i].nodeType == TEXT_NODE
                   && childs[i].nodeValue.search(/\S/) != -1
                   ){             
            countViews = 100; 
        }
        if (countViews >1)
            break;
    }
    
    if (countViews == 1) {
        return childView;
    }
    var containerConfig = document.createElementNS(WOBZILLA_NS,"container");
    for (var i = 0;i<childs.length; i++) {
        if (childs[i].nodeType == TEXT_NODE
            && childs[i].nodeValue.search(/\S/) == -1) {
            // ignore whitespace text-Nodes
            continue;
        }
        containerConfig.appendChild(childs[i].cloneNode(true));
    }
    return containerConfig;    
}

/** tests if the child-elements of a configuration are excactly one valid view entry 
or otherwise creates a WbzContainer Wrapper-Configuration.
 @return the child-configuration or if neccessary a configuration of a wrapper object
*/
function WBZ_wrapConfig(config) {
    var childs = config.childNodes;
    if (childs == null)
        throw new Error(node.nodeName + " not an element node ");
    return WBZ_wrapChildConfigs(childs);
}

/**
 returns the child elments (only the elements not the text-nodes)
 of an element-node.
 */
function WBZ_getElements(node) {
    var childs = node.childNodes;
    if (childs == null)
        throw new Error(node.nodeName + " not an element node ");
    var result = new Array();
    for (var i = 0;i<childs.length; i++) {
       if (childs[i].nodeType == ELEMENT_NODE)
          result.push(childs[i]);
    }
    return result;
}

/** returns the first child element (only the element not the text-nodes)
 of an element-node.
 */
function WBZ_getFirstElement(node) {
    var childs = node.childNodes;
    if (childs == null)
        throw new Error(node.nodeName + " not an element node ");
    for (var i = 0;i<childs.length; i++) {
       if (childs[i].nodeType == ELEMENT_NODE)
          return childs[i];
    }
    throw new Error(node.nodeName + " not an child element ");
}

/**
 returns the child nodes elments and text-nodes
 of an element-node. Use this method instead of
 node.childNodes when you are modifing th 
 node while iteratin childNodes.
 */
function WBZ_getChilds(node) {
    var childs = node.childNodes;
    var result = new Array();
    for (var i = 0;i<childs.length; i++) {
        result.push(childs[i]);
    }
    return result;
}

/**
 returns the child nodes elments and text-nodes
 of an element-node. Use this method instead of
 node.childNodes when you are modifing th 
 node while iteratin childNodes.
 */
function WBZ_getChildsByTagName(node,tagName) {
    var result = new Array();
    var childs = node.childNodes;
    for (var i = 0;i<childs.length; i++) {
        if (childs[i].nodeType != ELEMENT_NODE)
            continue;

        if ( childs[i].localName == tagName ) {
           result.push(childs[i]);
        }
    }
    return result;
}

function WBZ_isAncestorOf(node,childNode) {
    var currentNode = childNode.parentNode;
    while (currentNode != null) {
        if (node.isSameNode(currentNode))
            return true;
        currentNode = currentNode.parentNode;
    }
    return false
}

function WBZ_isAncestorOrSelf(node,childNode) {
    var currentNode = childNode;
    while (currentNode != null) {
        if (node.isSameNode(currentNode))
            return true;
        currentNode = currentNode.parentNode;
    }
    return false
}

/** Tests, if one of the changes events is targeted to childrens of
 * the model or the model itself. */
function WBZ_isTargeted(changes,model) {
    if (changes.localVariableChanged 
        || changes.globalVariableChanged 
        || changes.positionChanged 
        || changes.paramChanged) {
        return true;
    }       

    for (var i=0;i<changes.events.length;i++) {
        var event = changes.events[i];
        if (WBZ_isAncestorOrSelf(model,event.target))
            return true;
        if (event.type == "DOMNodeRemoved") {
            if (WBZ_isAncestorOrSelf(model,event.relatedNode)) {
                return true;
            }
        }

    }
    return false;
}

/** same as parentNode.removeChild(node).
  But the exception includes a stacktrace. */
function WBZ_remove(parentNode, node) {
    try {
        parentNode.removeChild(node);
    } catch (ex) {
        throw new Error(ex.message);
    }
}

/**
   Removes all child nodes from the node.
 */
function WBZ_removeAll(node) {
    try {
        var childs = WBZ_getChilds(node);;
        for (var i = 0;i<childs.length; i++) {
            node.removeChild(childs[i]);
        }
    } catch (ex) {
        throw new Error(ex.message);
    }
}

/**
   Insert <code>newNode</node> after <code>insertPoint</code>
   under the <code>parentNode</code>. If <code>insertPoint</code> is null,
   the node will be appended behind the last child-node of <code>parentNode</code>
 */
function WBZ_insertAfter(parentNode,newNode,insertPoint) {
    try {
        if (insertPoint != null) {
            parentNode.insertBefore(newNode,insertPoint.nextSibling);
        } else {
            parentNode.appendChild(newNode);
        }
    } catch (ex) {
        throw new Error(ex.message);
    }
}

/**Insert <code>newNode</node> before <code>insertPoint</code>
   under the <code>parentNode</code>. If <code>insertPoint</code> is null,
   the node will be appended behind the last child-node of <code>parentNode</code> */
function WBZ_insertBefore(parentNode,newNode,insertPoint) {
    try {
        parentNode.insertBefore(newNode,insertPoint);
    } catch (ex) {
        throw new Error(ex.message);
    }
}

/** Test if the two list are equal @param equals a function that test equality of elements in listA with elements of listB
@return true if lists are equal
*/
function WBZ_equals(listA,listB,equals) {
    if (listA.length != listB.length) {
        return false;
    } else {
        for (var i=0;i<listA.length;i++) {
            if (!equals(listA[i],listB[i])) {
                return false;
            }
        }
    }
    return true;
}

var MATCH_STATE = 0;
var TEST_A_STATE = 1;
var TEST_B_STATE = 2;
/** Performes a diff between to lists.
   A variant of the classical VMS-DIFF Algorithm, that
   doesn't use the LCS approach. Worst-case is 
   quadratic run-time when the lists have common elements.
   @param listA the first list as Array
   @param listB the second list as Array
   @param diffA an Array to store the elements that are inserted into A
   @param diffB an Array to store the elements that are removed from A
   @param equals a function that test equality of elements in listA with elements of listB
 */
function WBZ_diff(listA,listB,diffA,diffB,equals)
{
    var i = 0; 
    var n = listA.length;
    var j = 0;
    var m = listB.length;
    var bufA = new Array();
    var bufB = new Array();
    var state = MATCH_STATE;
    while (true) {
        //WBZ_info("STATE " + state + " i " + i + " j " + j);

        // Test if list-borders are reached
        if ((i == n && state != TEST_B_STATE) 
            || (j == m && state != TEST_A_STATE) 
            )
        {
            if ( i == n) {
                var element = bufB.shift();
                while (element != null) {
                    diffB.push(element);
                    element = bufB.shift();
                }
                var last = (i>0) ? listA[i-1] : null;                    
                for (var k=j;k<m;k++) {
                    if (last == null || !equals(last,listB[k]))
                        diffB.push(listB[k]);
                }
            } 
            if ( j == m) {
                var element = bufA.shift();
                while (element != null) {
                    diffA.push(element);
                    element = bufA.shift();
                }
                var last = (j>0) ? listB[j-1] : null;                    
                for (var k=i;k<n;k++) {
                    if (last == null || !equals(listA[k],last))
                        diffA.push(listA[k]);
                }
            }
            return diffA.length + diffB.length;
        }


        switch (state) {
        case MATCH_STATE: //Read and compare on element of each list 
            if (!equals(listA[i],listB[j])) {
                // push to stackA and stackB
                bufA.push(listA[i]);
                bufB.push(listB[j]);
                state = TEST_A_STATE;
            }
            i++;
            j++;
            break;
        case TEST_A_STATE: //Read element from the first list
            var match = false;
            // compare with all elements from stackB
            for (var k=0;k<bufB.length;k++) { 
                if (equals(listA[i],bufB[k])) {
                    match = true;
                    // all elements from StackA can't 
                    // be matched, so add them to the diff
                    var element = bufA.shift();
                    while (element != null) {
                        //WBZ_info("pushing " + element.getAttribute("name"));
                        diffA.push(element);
                        element = bufA.shift();
                    }
                    var length = bufB.length;
                    // remove the entries following the matched entry
                    for (var l=k;l<length;l++) {
                        bufB.pop();
                    }
                    j = j - (length- k -1);
                    break;
                }
            }
            if (!match) {
                bufA.push(listA[i]); // 
                state = TEST_B_STATE;
            } else {
                state = MATCH_STATE;
            }
            i++;
            break;
        case TEST_B_STATE:
            var match = false;
            // compare with all elements from stackA
            for (var k=0;k<bufA.length;k++) {
                if (equals(bufA[k],listB[j])) {
                    match = true;
                    // all elements from StackB can't 
                    // be matched, so add them to the diff
                    var element = bufB.shift();
                    while (element != null) {
                        //WBZ_info("pushing " + element.model.getAttribute("name"));
                        diffB.push(element);
                        element = bufB.shift();
                    }
                    var length = bufA.length; 
                    // remove the entries following the matched entry
                    for (var l=k;l<length;l++) {
                        bufA.pop();
                    }
                    i = i - (length- k -1);
                    break; 
                } else {
                }
            }
            if (!match) {
                bufB.push(listB[j]);
                state = TEST_A_STATE;
            } else {
                state = MATCH_STATE;
            }
            j++;
            break;
        }
    }

    // Idea: Should the result-diff contain additional information like: diffA = {position,n,element_1,...element_n,position,...}
}


/** creates a new object and copies all properties 
    and method from the original.
    @returns the clone
 */
function WBZ_clone(original) {
    var clone = new Object();
    for (var property in original) {
        clone[property] = original[property];
    }
    return clone;
}


/** Set the global debuggin level*/
function WBZ_setDebuggingLevel(level) {
    VERBOSE_LEVEL = level;
    DEBUGGING = (VERBOSE_LEVEL <= DEBUG_LEVEL);
}

var WBZ_staticDebugger = null;
/** You can debug into a global logger-element with the id log_window */
function WBZ_getDebugger(message) {
    if (WBZ_staticDebugger == null) {
        var outputNode = document.getElementById('log_window');
        if (outputNode == null)
            throw new Error("Can't find element with id 'log_window'");
        
        WBZ_staticDebugger = new DefaultLogger("static", outputNode);
    }
    return WBZ_staticDebugger;
}

/** uses global Logger-Object for debugging */
function WBZ_debug(message) {
    WBZ_getDebugger().debug(message);
}

/** uses global Logger-Object for debugging */
function WBZ_info(message) {
    WBZ_getDebugger().info(message);
}

function assertNotNull(arg,text) {
    if (text == null)
        text = "";
    if (arg == null) {
        throw new Error("Not Null Assertion failed: " + text);
    }
}

function WBZ_dumpProperties(element) {
  var message = "\n";
  for (var list in element) {
     message += list+"\n";
  }
  WBZ_debug(message);
}



/*
var WOBZILLA_id;

function WBZ_removeId(element,recursive) {
    if (recursive) {
        WBZ_removeId(element);
        var childs = element.childNodes;
        for (var i=0;i<childs.length;i++) {
            if (childs[i].nodeType == ELEMENT_NODE)
                WBZ_removeId(childs[i],true);
        }
    } else {
        element.removeAttributeNS(WOBZILLA_NS,"id");
    }
}

function WBZ_setNewId(element,recursive)
{
    if (recursive) {
        var newId = WBZ_setNewId(element);
        var childs = element.childNodes;
        for (var i=0;i<childs.length;i++) {
            if (childs[i].nodeType == ELEMENT_NODE)
                WBZ_setNewId(childs[i],true);
        }
        return newId;
    } else {
        if (WOBZILLA_id == null)
            WOBZILLA_id = 0;
        WOBZILLA_id ++;
        //        WBZ_debug("Create id for " + element.nodeName);
        element.setAttributeNS(WOBZILLA_NS,"wbz:id",WOBZILLA_id);
        return WOBZILLA_id;
    }
}
*/

